import gai.common.ConfigHelper as ConfigHelper
from gai.common.http_utils import http_post
API_BASEURL = ConfigHelper.get_api_baseurl()


class STTClient:

    def __call__(self, generator=None, file=None):
        if generator == "openai-whisper":
            return self.openai_whisper(file=file)

        if not file:
            raise Exception("No file provided")

        files = {
            "model": (None, generator),
            "file": (file.name, file.read())
        }

        response = http_post(
            f"{API_BASEURL}/gen/v1/audio/transcriptions", files=files)
        response.decode = lambda: response.json()["text"]
        return response

    def openai_whisper(self, **model_params):
        import os
        import openai
        from openai import OpenAI
        from dotenv import load_dotenv
        load_dotenv()
        if not os.environ.get("OPENAI_API_KEY"):
            raise Exception(
                "OPENAI_API_KEY not found in environment variables")
        openai.api_key = os.environ["OPENAI_API_KEY"]
        client = OpenAI()

        if "file" not in model_params:
            raise Exception("Missing file parameter")

        file = model_params["file"]

        # If file is a bytes object, we need to write it to a temporary file then pass the file object to the API
        if isinstance(file, bytes):
            import tempfile
            with tempfile.NamedTemporaryFile(suffix='.wav') as temp:
                temp.write(file)
                temp.flush()
                temp.seek(0)
                model_params["file"] = temp.file
                response = client.audio.transcriptions.create(
                    model='whisper-1', **model_params)
        else:
            response = client.audio.transcriptions.create(
                model='whisper-1', **model_params)

        return response



import json
class OpenAIChunkWrapper:

    # chunk is json in binary representation
    def __init__(self, chunk):
        self.chunk = chunk

    def decode(self):
        return self.chunk.choices[0].delta.content
    




class HttpException(Exception):
    def __init__(self, status_code, code, message, url):
        self.message = message
        self.status_code = status_code
        self.code = code
        self.url = url
        super().__init__(self.message)
    pass


import os
import sys
import re
import time
import json
from os.path import dirname
import shutil

# Create ~/.gaiaio/cache

# Create ~/.cache/locallm


def mkdir_cache(sub_dir):
    cache_dir = os.path.join(os.path.join(
        os.path.expanduser('~'), '~/gai/cache/'), sub_dir)
    os.makedirs(cache_dir, exist_ok=True)
    return cache_dir


def is_url(s):
    return re.match(r'^https?:\/\/.*[\r\n]*', s) is not None


def sha256(text):
    import hashlib
    return hashlib.sha256(text.encode('utf-8')).hexdigest()


def timestamp():
    return int(time.time() * 1000)


def find_url_in_text(text):
    url_pattern = re.compile(
        r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')
    urls = re.findall(url_pattern, text)
    return urls

# Remove all white spaces and quotation marks from a string for use as a prompt


def clean_string(s):
    if s is None:
        return ''
    s = re.sub(r'[\s\'\"]+', ' ', s)
    s = s.replace('\\"', '"').replace('\\n', '')
    return s.strip()

# Remove unprintable unicode chars


def remove_unprintable(input_str):
    return ''.join(ch for ch in input_str if ch.isprintable() or ch in ('\t', '\n'))

# This is useful when asking for JSON results from LLM


def extract_json_substring(s):
    s = re.sub(r"\s+", " ", s)
    s = s.replace('\\"', '"').replace('\\n', '')
    try:
        match = re.search(r'\[.*\]', s)
        if match:
            json_string = match.group(0)
            return json_string
    except:
        pass
    return None


def extract_double_quoted_substring(s):
    match = re.search(r'\\"(.*)\\"', s)
    if match:
        # This will hold 'What is the original purpose of the Apple Remote?'
        result = match.group(1)
        return result
    return None


def find_site_packages_path(virtual_env_name):
    # Extracting the Python version
    python_version = f"{sys.version_info.major}.{sys.version_info.minor}"

    # Constructing the path pattern to the virtual environment's site-packages directory
    site_packages_path = os.path.expanduser(
        f"~/miniconda/envs/{virtual_env_name}/lib/python{python_version}/site-packages")
    return site_packages_path


def find_egg_link(virtual_env_name, package_name):
    site_packages_path = find_site_packages_path(virtual_env_name)
    egg_link_file = os.path.join(
        site_packages_path, f"{package_name}.egg-link")
    if os.path.exists(egg_link_file):
        return egg_link_file
    else:
        return None


def find_project_path(virtual_env_name, package_name):
    egg_link_file = find_egg_link(virtual_env_name, package_name)
    if egg_link_file is None:
        return None
    else:
        with open(egg_link_file) as f:
            project_path = f.readline().strip()
            return project_path


def is_data_url(s):
    """
    Check if the given string is a valid data URL.

    Args:
        s (str): The string to check.

    Returns:
        bool: True if the string is a valid data URL, False otherwise.
    """
    data_url_pattern = re.compile(r'data:[\w\-]+\/[\w\-]+;base64,[\w\+=/]+')
    return bool(data_url_pattern.match(s))


def is_image_data_url(s):
    """
    Check if the given string is a valid data URL.

    Args:
        s (str): The string to check.

    Returns:
        bool: True if the string is a valid data URL, False otherwise.
    """
    data_url_pattern = re.compile(r'data:image\/[\w\-]+;base64,[\w\+=/]+')
    return bool(data_url_pattern.match(s))



import gai.common.ConfigHelper as ConfigHelper
from gai.common.http_utils import http_post
API_BASEURL = ConfigHelper.get_api_baseurl()
import json, requests

class RAGClient:

    def index_file(self, collection_name, file_path, metadata={"source":"unknown"}):
        with open(file_path, "r") as f:
            url =  f"{API_BASEURL}/gen/v1/rag/index_file"
            headers = {
                'accept': 'application/json',
            }
            data={
                "collection_name":collection_name,
                "metadata":json.dumps(metadata)
            }
            files={
                "file":f, 
            }
            response = requests.post(url=url,headers=headers,data=data,files=files)
            return response

    def retrieve(self, collection_name, query_texts, n_results=None):
        data={
            "collection_name":collection_name,
            "query_texts":query_texts
        }
        if n_results:
            data["n_results"]=n_results
       
        response = http_post(f"{API_BASEURL}/gen/v1/rag/retrieve",data=data)
        return response





from gai.lib.ttt.ChunkWrapper import ChunkWrapper
from gai.lib.ttt.OpenAIChunkWrapper import OpenAIChunkWrapper
from gai.lib.ttt.AnthropicChunkWrapper import AnthropicChunkWrapper
import gai.common.ConfigHelper as ConfigHelper
from gai.common.http_utils import http_post
from gai.common.generators_utils import chat_string_to_list, chat_list_to_string
from gai.common._exceptions import HttpException
import json

API_BASEURL = ConfigHelper.get_api_baseurl()


class TTTClient:

    def api(self, generator=None, messages=None, stream=True, **generator_params):
        if not messages:
            raise Exception("Messages not provided")

        if isinstance(messages, str):
            messages = chat_string_to_list(messages)

        cli_config = ConfigHelper.get_cli_config()
        if not generator:
            generator = cli_config["default_generator"]

        data = {
            "model": generator,
            "messages": messages,
            "stream": stream,
            **cli_config["generators"][generator]["default"],
            **generator_params
        }

        def streamer(response):
            for chunk in response.iter_lines():
                yield ChunkWrapper(chunk)

        try:
            cli_baseurl = cli_config["generators"][generator]["url"]
            response = http_post(f"{cli_baseurl}/completions", data)
        except HttpException as he:
            if he.code == "context_length_exceeded":
                try:
                    data["model"] = "mistral7b_128k-exllama"
                    generator = "mistral7b_128k-exllama"
                    cli_baseurl = cli_config["generators"][generator]["url"]
                    response = http_post(f"{cli_baseurl}/completions", data)
                except Exception as e:
                    raise e
            else:
                raise he
        except Exception as e:
            raise e

        if not stream:
            response.decode = lambda: response.json(
            )["choices"][0]["message"]["content"]
            return response
        return streamer(response)

    def __call__(self, generator=None, messages=None, stream=True, **generator_params):
        if generator == "gpt-4":
            return self.gpt_4(messages=messages, stream=stream, **generator_params)
        elif generator == "claude2-100k":
            return self.claude_2(messages=messages, stream=stream, **generator_params)
        return self.api(generator, messages, stream, **generator_params)

    def gpt_4(self, messages=None, stream=True, **generator_params):
        import os
        import openai
        from openai import OpenAI
        from dotenv import load_dotenv
        load_dotenv()
        if not os.environ.get("OPENAI_API_KEY"):
            raise Exception(
                "OPENAI_API_KEY not found in environment variables")
        openai.api_key = os.environ["OPENAI_API_KEY"]
        client = OpenAI()

        if not messages:
            raise Exception("Messages not provided")

        def streamer(response):
            for chunk in response:
                yield OpenAIChunkWrapper(chunk)

        model = "gpt-4"
        if isinstance(messages, str):
            messages = chat_string_to_list(messages)
        response = client.chat.completions.create(
            model=model,
            messages=messages,
            stream=stream,
            **generator_params
        )

        if not stream:
            response.decode = lambda: response.choices[
                0].message.content if response.choices[0].message.content else ""
            return response
        return streamer(response)

    def claude_2(self, messages=None, stream=True, **generator_params):
        import os
        from anthropic import Anthropic
        from dotenv import load_dotenv
        load_dotenv()
        if not os.environ.get("ANTHROPIC_API_KEY"):
            raise Exception(
                "ANTHROPIC_API_KEY not found in environment variables")
        client = Anthropic()

        if not messages:
            raise Exception("Messages not provided")

        def streamer(response):
            for chunk in response:
                yield AnthropicChunkWrapper(chunk)

        model = "claude-2"
        message = messages
        if isinstance(messages, list):
            message = chat_list_to_string(messages)
        prompt_template = "\n\nHuman: {message}\n\nAssistant:"
        messages = prompt_template.format(message=message)
        response = client.completions.create(
            model=model,
            prompt=messages,
            stream=stream,
            **generator_params
        )

        if not stream:
            response.decode = lambda: response.completion
            return response
        return streamer(response)



import re
import json
from gai.common.ConfigHelper import get_cli_config

# A simple utility to validate if all items in model params are in the whitelist.


def validate_params(model_params, whitelist_params):
    for key in model_params:
        if key not in whitelist_params:
            raise Exception(
                f"Invalid param '{key}'. Valid params are: {whitelist_params}")

# A simple utility to filter items in model params that are also in the whitelist.


def filter_params(model_params, whitelist_params):
    filtered_params = {}
    for key in model_params:
        if key in whitelist_params:
            filtered_params[key] = model_params[key]
    return filtered_params

# A simple utility to load generators config.


def load_generators_config():
    return get_cli_config()["gen"]

# This is useful for converting chatgpt-style dialog to text dialog


def chat_list_to_string(messages):
    if type(messages) is str:
        return messages
    prompt = ""
    for message in messages:
        content = message['content'].strip()
        role = message['role'].strip()
        if content:
            prompt += f"{role}: {content}\n"
        else:
            prompt += f"{role}:"
    return prompt

# This is useful for converting text dialog to chatgpt-style dialog


def chat_string_to_list(messages, ai_name="assistant"):
    # Split the messages into lines
    lines = messages.split('\n')

    # Prepare the result list
    result = []

    # Define roles
    roles = ['system', 'user', ai_name]

    # Initialize current role and content
    current_role = None
    current_content = ''

    # Process each line
    for line in lines:
        # Check if the line starts with a role
        for role in roles:
            # ignore case when comparing roles
            if line.lower().startswith(role.lower() + ':'):
                # If there is any content for the current role, add it to the result
                if current_role is not None and current_content.strip() != '':
                    result.append(
                        {'role': current_role, 'content': current_content.strip()})

                # Start a new role and content
                current_role = role
                current_content = line[len(role) + 1:].strip()
                break
        else:
            # If the line does not start with a role, add it to the current content
            current_content += ' ' + line.strip()

    # Add the last role and content to the result
    if current_role is not None:
        result.append(
            {'role': current_role, 'content': current_content.strip()})

    return result


# This is the async version of the word_streamer
async def word_streamer_async(char_generator):
    buffer = ""
    async for byte_chunk in char_generator:
        if type(byte_chunk) == bytes:
            byte_chunk = byte_chunk.decode("utf-8", "replace")
        buffer += byte_chunk
        words = buffer.split(" ")
        if len(words) > 1:
            for word in words[:-1]:
                yield word
                yield " "
            buffer = words[-1]
    yield buffer

# This is useful for converting a stream of characters to a stream of words


def word_streamer(char_generator):
    buffer = ""
    for chunk in char_generator:
        if chunk:
            if type(chunk) == bytes:
                chunk = chunk.decode("utf-8", "replace")
            buffer += chunk
            words = buffer.split(" ")
            if len(words) > 1:
                for word in words[:-1]:
                    yield word
                    yield " "
                buffer = words[-1]
    yield buffer


# This is useful for accepting a stream of string and converting it to a stream of JSON objects
def json_streamer(httpResponse):
    httpResponse.encoding = 'utf-8'
    for line in httpResponse.iter_lines(decode_unicode=True):
        if line:
            yield json.loads(str(line))


# This is useful for extracting argument when expecting the AI to return a function prompt
def extract_argument_from_prompt_function(output, command, remove_space=True, remove_quotes=True, remove_slash=True, remove_dollar=True):
    # Strip whitespaces, double quotes, single quotes, and backslashes
    if (remove_space):
        output = re.sub(r'\s+', '', output)
        command = re.sub(r'\s+', '', command)

    if (remove_quotes):
        output = re.sub(r'"|\'', '', output)
        command = re.sub(r'"|\'', '', command)

    if (remove_slash):
        output = re.sub(r'\\+', '', output)
        command = re.sub(r'\\+', '', command)

    if (remove_dollar):
        output = re.sub(r'\\$', '', output)
        command = re.sub(r'\\$', '', command)

    # Capture argument after "=" or within "()"
    print("cleaned_command=", command)
    print("output_command=", output)
    query_string = output.replace(command, '').strip().lstrip(
        '(').rstrip(')').lstrip('=').strip()
    return query_string

    # matches = re.search(rf"^{re.escape(command)}(?:\s*=\s*|\s*\(\s*)\"?([^\)\"\']+)\"?(?:\s*\)|\b)", output)
    # if matches:
    #     query_string = matches.group(1)
    #     return query_string



import json
class ChunkWrapper:

    # chunk is json in binary representation
    def __init__(self, chunk):
        self.chunk = chunk

    def __str__(self):
        return self.chunk.decode('utf-8')

    def decode(self):
        decoded_chunk = json.loads(self.chunk.decode('utf-8'))
        if "content" in decoded_chunk["choices"][0]["delta"] and decoded_chunk["choices"][0]["delta"]['content']:
            return decoded_chunk["choices"][0]["delta"]["content"]
        if decoded_chunk["choices"][0]["delta"]['tool_calls'] and decoded_chunk["choices"][0]["delta"]['tool_calls'][0]['function']['name']:
            return {"tool_name": decoded_chunk["choices"][0]["delta"]['tool_calls'][0]['function']['name']}, "tool_name"
        if decoded_chunk["choices"][0]["delta"]['tool_calls'] and decoded_chunk["choices"][0]["delta"]['tool_calls'][0]['function']['arguments']:
            return {"tool_arg": decoded_chunk["choices"][0]["delta"]['tool_calls'][0]['function']['arguments']}, "tool_arg"
        if decoded_chunk["choices"][0]["finish_reason"]:
            return {"finish_reason": decoded_chunk["choices"][0]["finish_reason"]}, "finish_reason"
        return ""
    
    def __dict__(self):
        return json.loads(self.chunk.decode('utf-8'))




import os
import json
from os.path import dirname
import shutil


def init():
    with open(os.path.expanduser("~/.gairc"), "w") as file:
        file.write(json.dumps({
            "app_dir": "~/gai",
            "discovery_url": "",
        }, indent=4))


def get_rc():
    if (not os.path.exists(os.path.expanduser("~/.gairc"))):
        init()
    return json.load(open(os.path.expanduser("~/.gairc")))


def get_api_baseurl():
    return "https://gaiaio.ai/api"


def get_cli_config():
    rc = get_rc()
    app_dir = rc["app_dir"]
    cli_config = os.path.join(os.path.expanduser(app_dir), "cli-config.json")

    if not os.path.exists(cli_config):
        cli_config = {
            "default_generator": "mistral7b-exllama",
            "generators": {
                "mistral7b-exllama": {
                    "type": "ttt",
                    "url": "https://gaiaio.ai/api/gen/v1/chat",
                    "whitelist": [
                        "temperature",
                        "top_p",
                        "min_p",
                        "top_k",
                        "max_new_tokens",
                        "typical",
                        "n",
                        "token_repetition_penalty_max",
                        "token_repetition_penalty_sustain",
                        "token_repetition_penalty_decay",
                        "beams",
                        "beam_length"
                    ],
                    "default": {
                        "temperature": 1.2,
                        "top_p": 0.15,
                        "min_p": 0.0,
                        "top_k": 50,
                        "max_new_tokens": 1000,
                        "typical": 0.0,
                        "token_repetition_penalty_max": 1.25,
                        "token_repetition_penalty_sustain": 256,
                        "token_repetition_penalty_decay": 128,
                        "beams": 1,
                        "beam_length": 1
                    }
                },
                "mistral7b_128k-exllama": {
                    "type": "ttt",
                    "url": "https://gaiaio.ai/api/gen/v1/longchat",
                    "whitelist": [
                        "temperature",
                        "top_p",
                        "min_p",
                        "top_k",
                        "max_new_tokens",
                        "typical",
                        "n",
                        "token_repetition_penalty_max",
                        "token_repetition_penalty_sustain",
                        "token_repetition_penalty_decay",
                        "beams",
                        "beam_length"
                    ],
                    "default": {
                        "temperature": 1.2,
                        "top_p": 0.15,
                        "min_p": 0.0,
                        "top_k": 50,
                        "max_new_tokens": 1000,
                        "typical": 0.0,
                        "token_repetition_penalty_max": 1.25,
                        "token_repetition_penalty_sustain": 256,
                        "token_repetition_penalty_decay": 128,
                        "beams": 1,
                        "beam_length": 1
                    }
                },
                "gpt-4": {
                    "type": "ttt",
                    "whitelist": [
                        "max_tokens",
                        "temperature",
                        "top_p",
                        "presence_penalty",
                        "frequency_penalty",
                        "stop",
                        "logit_bias",
                        "n",
                        "stream",
                        "openai_api_key"
                    ],
                    "default": {
                        "temperature": 1.2,
                        "top_p": 0.15,
                        "top_k": 50,
                        "max_tokens": 1000
                    }
                },
                "claude2-100k": {
                    "type": "ttt",
                    "whitelist": [
                        "max_tokens_to_sample",
                        "temperature",
                        "top_p",
                        "top_k",
                        "stop_sequences"
                    ],
                    "default": {
                        "temperature": 1.2,
                        "top_p": 0.15,
                        "top_k": 50,
                        "max_tokens_to_sample": 1000
                    }
                }
            },
        }

    return cli_config


def get_default_generator():
    cli_config = get_cli_config()
    return cli_config["default_generator"]


def get_generator_url(generator):
    cli_config_json = get_cli_config()
    if generator not in cli_config_json['generators']:
        raise Exception(f"Generator {generator} not found supported.")
    return cli_config_json['generators'][generator]["url"]



import re
import httpx,requests,json
from gai.common.logging import logging
logger = logging.getLogger(__name__)
import pprint
import os
from urllib.parse import urlparse
from gai.common._exceptions import HttpException

def is_url(s):
    return re.match(r'^https?:\/\/.*[\r\n]*', s) is not None

# Check if URL contains a file extension (e.g. .pdf, .jpg, .png, etc.)
def has_extension(url):
    parsed_url = urlparse(url)
    _, ext = os.path.splitext(parsed_url.path)
    return bool(ext)                

async def http_post_async(url,data):
    return httppost_async(url,data)    

async def httppost_async(url, data):
    headers = {"Content-Type": "application/json"}
    async with httpx.AsyncClient() as client:
        if not isinstance(data, str):
            data = json.dumps(data)
        try:
            response = await client.post(url, data=data, headers=headers)
            if response.status == 200:
                return response
            else:
                try:
                    error_object = response.json()
                    code = error_object.get("code", "unknown")
                    message = error_object.get("message", "unknown")
                except Exception as e2:
                    code = "unknown"
                    message = response.text()
                raise HttpException(response.status_code, code, message, url)
        except httpx.HTTPStatusError as e:
            raise Exception("Connection Error. Is the service Running?")

def http_post(url, data=None, files=None):
    return httppost(url,data,files)    

def httppost(url, data=None, files=None):
    if data==None and files==None:
        raise Exception("No data or files provided")

    logger.debug(f"httppost:url={url}") 
    logger.debug(f"httppost:data={pprint.pformat(data)}") 
    headers = {"Content-Type": "application/json"}
    try:
        if files:
            if data and "stream" in data:
                files["stream"] = (None, data["stream"])
            response = requests.post(url, files=files)
        else:
            if "stream" in data:
                response = requests.post(url, json=data, headers=headers, stream=data["stream"])
            else:
                response = requests.post(url, json=data, headers=headers)
        if response.status_code == 200:
            return response
        else:
            try:
                error_object = response.json()
                code = error_object.get("code", "unknown")
                if (code == "unknown"):
                    code = response.reason.lower().replace(' ', '_')
                message = str(error_object)
                if (message == "unknown"):
                    message = response.text
            except Exception as e2:
                if response.reason:
                    code = response.reason.lower().replace(' ', '_')
                    message = response.reason
                else:
                    code = response.status_code
                    message = response.text
            raise HttpException(response.status_code, code, message, url)

    except requests.exceptions.ConnectionError as e:
        raise Exception("Connection Error. Is the service Running?")

def http_get(url):
    return httpget(url)

def httpget(url):
    try:
        response = requests.get(url)
        if response.status_code == 200:
            return response
        else:
            raise Exception(f"httpget: Error={response.text}")
    except requests.exceptions.ConnectionError as e:
        raise Exception("Connection Error. Is the service Running?")

async def http_get_async(url):
    return httpget_async(url)

async def httpget_async(url):
    async with httpx.AsyncClient() as session:
        try:
            async with session.get(url) as response:
                if response.status == 200:
                    return response                      # Returning the data
                else:
                    raise Exception(f"httpget_async: Error={await response.text()}")
        except httpx.HTTPStatusError as e:
            raise Exception("Connection Error. Is the service Running?")

def http_delete(url):
    return httpdelete(url)

def httpdelete(url):
    try:
        response = requests.delete(url)
        if response.status_code == 200:
            return response
        else:
            raise Exception(f"httpdelete: Error={response.text}")
    except requests.exceptions.ConnectionError as e:
        raise Exception("Connection Error. Is the service Running?")

def http_put(url):
    return httpput(url)

def httpput(url):
    try:
        response = requests.put(url)
        if response.status_code == 200:
            return response
        else:
            raise Exception(f"httpput: Error={response.text}")
    except requests.exceptions.ConnectionError as e:
        raise Exception("Connection Error. Is the service Running?")


    


from unstructured.partition.pdf import partition_pdf
import re

class PDFConvert:

    @staticmethod
    def pdf_to_text(pdf_file_path, clean=True):
        elements = partition_pdf(pdf_file_path)
        text="\n\n".join([str(el) for el in elements])
        text = re.sub(r'\(cid:[^\)]*\)', '', text)
        if clean:
            text = re.sub(r'\s+', ' ', text)
        return text


